import {shallowCompareKeys} from '@blueprintjs/core/lib/cjs/common/utils';
import React from 'react';

import {filterByQuery} from '../app/GraphQueryImpl';
import {GanttChartLayout} from '../gantt/Constants';
import {GanttChartMode} from '../gantt/GanttChart';
import {buildLayout} from '../gantt/GanttChartLayout';
import {explodeCompositesInHandleGraph} from '../pipelines/CompositeSupport';
import {StepEventStatus} from '../types/globalTypes';

import {PartitionRunMatrixPipelineQuery_pipelineSnapshotOrError_PipelineSnapshot_solidHandles} from './types/PartitionRunMatrixPipelineQuery';
import {PartitionRunMatrixRunFragment} from './types/PartitionRunMatrixRunFragment';
import {PartitionRuns} from './useChunkedPartitionsQuery';

type SolidHandle = PartitionRunMatrixPipelineQuery_pipelineSnapshotOrError_PipelineSnapshot_solidHandles;
type StatusSquareColor =
  | 'SUCCESS'
  | 'FAILURE'
  | 'SKIPPED'
  | 'MISSING'
  | 'FAILURE-SUCCESS'
  | 'FAILURE-SKIPPED'
  | 'SUCCESS-FAILURE'
  | 'SUCCESS-SKIPPED'
  | 'SKIPPED-SUCCESS'
  | 'SKIPPED-FAILURE';

export const StatusSquareFinalColor: {[key: string]: StatusSquareColor} = {
  'FAILURE-SUCCESS': 'SUCCESS',
  'SKIPPED-SUCCESS': 'SUCCESS',
  'SUCCESS-FAILURE': 'FAILURE',
  'SKIPPED-FAILURE': 'FAILURE',
  'FAILURE-SKIPPED': 'SKIPPED',
  'SUCCESS-SKIPPED': 'SKIPPED',
};

export interface DisplayOptions {
  showFailuresAndGapsOnly: boolean;
  showPrevious: boolean;
  colorizeByAge: boolean;
  colorizeSliceUnix: number;
}

const DYNAMIC_STEP_REGEX_SUFFIX = '\\[.*\\]';

export interface MatrixStep {
  name: string;
  color: string;
  unix: number;
}

function getStartTime(a: PartitionRunMatrixRunFragment) {
  return ('startTime' in a.stats && a.stats.startTime) || 0;
}

function byStartTimeAsc(a: PartitionRunMatrixRunFragment, b: PartitionRunMatrixRunFragment) {
  return getStartTime(a) - getStartTime(b);
}

// BG Note: Dagit 0.10.0 removed the .compute step key suffix, but the Run Matrix takes the current
// step tree and looks up data for each step in historical runs. For continuity across 0.10.0, we
// match historical step keys with the .compute format as well. We can remove safely after 120 days?
export function isStepKeyForNode(nodeName: string, stepKey: string) {
  const dynamicRegex = new RegExp(nodeName + DYNAMIC_STEP_REGEX_SUFFIX);
  return stepKey === nodeName || stepKey === `${nodeName}.compute` || stepKey.match(dynamicRegex);
}

function buildMatrixData(
  layout: GanttChartLayout,
  partitions: PartitionRuns[],
  options: DisplayOptions,
) {
  // Note this is sorting partition runs in place, I don't think it matters and
  // seems better than cloning all the arrays.
  partitions.forEach((p) => p.runs.sort(byStartTimeAsc));

  const partitionColumns = partitions.map((p) => ({
    ...p,
    steps: layout.boxes.map(({node}) => {
      const datapoints = p.runs
        .map((r, idx) => ({
          runIdx: idx,
          status: r.stepStats.find((stats) => isStepKeyForNode(node.name, stats.stepKey))?.status,
        }))
        .filter(
          (s): s is {runIdx: number; status: StepEventStatus} =>
            !!s.status && s.status !== StepEventStatus.IN_PROGRESS,
        )
        .reverse();
      if (datapoints.length === 0) {
        return {
          name: node.name,
          color: 'MISSING' as StatusSquareColor,
          unix: 0,
        };
      }

      // Calculate the box color for this step. CSS classes are in the "previous-final" format, and we'll
      // strip the "previous" half later if the user has that display option disabled.
      //
      // Rules:
      // - The `final` status is the status of the step the last time it was run
      // - The `previous` status is the status of the step before that run, if it was different.
      const prev = datapoints.slice(1).find((dp) => dp.status !== datapoints[0].status);
      const color = prev
        ? (`${prev.status}-${datapoints[0].status}` as StatusSquareColor)
        : datapoints[0].status;

      return {
        name: node.name,
        unix: getStartTime(p.runs[datapoints[0].runIdx]),
        color,
      };
    }),
  }));

  const partitionsWithARun = partitionColumns.filter((p) => p.runs.length > 0).length;

  const stepRows = layout.boxes.map((box, idx) => {
    const totalFailures = partitionColumns.filter((p) => p.steps[idx].color.includes('FAILURE'));
    const finalFailures = partitionColumns.filter((p) => p.steps[idx].color.endsWith('FAILURE'));
    return {
      x: box.x,
      name: box.node.name,
      totalFailurePercent: partitionsWithARun
        ? Math.round((totalFailures.length / partitionsWithARun) * 100)
        : 0,
      finalFailurePercent: partitionsWithARun
        ? Math.round((finalFailures.length / partitionsWithARun) * 100)
        : 0,
    };
  });

  if (options.showFailuresAndGapsOnly) {
    for (let ii = stepRows.length - 1; ii >= 0; ii--) {
      if (stepRows[ii].finalFailurePercent === 0) {
        stepRows.splice(ii, 1);
        partitionColumns.forEach((p) => p.steps.splice(ii, 1));
      }
    }
    for (let ii = partitionColumns.length - 1; ii >= 0; ii--) {
      if (
        partitionColumns[ii].runs.length === 0 ||
        partitionColumns[ii].steps.every((step) => step.color.includes('SUCCESS'))
      ) {
        partitionColumns.splice(ii, 1);
      }
    }
  }

  return {stepRows, partitions, partitionColumns};
}

interface MatrixDataInputs {
  solidHandles: SolidHandle[] | false;
  partitions: PartitionRuns[];
  stepQuery: string;
  options: DisplayOptions;
}

/**
 * This hook uses the inputs provided to filter the data displayed and calls through to buildMatrixData.
 * It uses a React ref to cache the result and avoids re-computing when all inputs are shallow-equal.
 *
 * - This could alternatively be implemented via React.memo and an outer + inner component pair, but I
 *   didn't want to split <PartitionRunMatrix />
 * - This can't be a React useEffect with an array of deps because we want the cached value to be updated
 *   synchronously when the inputs are modified to avoid a double-render caused by an effect + state var.
 *
 * @param inputs
 */
export const useMatrixData = (inputs: MatrixDataInputs) => {
  const cachedMatrixData = React.useRef<{
    result: ReturnType<typeof buildMatrixData>;
    inputs: MatrixDataInputs;
  }>();
  if (!inputs.solidHandles) {
    return null;
  }
  if (cachedMatrixData.current && shallowCompareKeys(inputs, cachedMatrixData.current.inputs)) {
    return cachedMatrixData.current.result;
  }

  const nodes = explodeCompositesInHandleGraph(inputs.solidHandles).map((h) => h.solid);

  // Filter the pipeline's structure and build the flat gantt layout for the left hand side
  const solidsFiltered = filterByQuery(nodes, inputs.stepQuery);

  const layout = buildLayout({nodes: solidsFiltered.all, mode: GanttChartMode.FLAT});

  // Build the matrix of step + partition squares - presorted to match the gantt layout
  const result = buildMatrixData(layout, inputs.partitions, inputs.options);
  cachedMatrixData.current = {result, inputs};
  return result;
};
